@@ -25,12 +25,10 @@ BEGIN {
     if ( $OSNAME eq 'MSWin32' ) {
         require Win32;
         require Win32::Process;
-        require Win32::Process::Info;
-        Win32::Process::Info->import();
     }
 }
 
-our $VERSION = '0.23';
+our $VERSION = '0.24';
 
 sub _ANYPROCESS                    { return -1 }
 sub _COMMAND                       { return 0 }
@@ -38,6 +36,7 @@ sub _DEFAULT_HOST                  { return 'localhost' }
 sub _WIN32_ERROR_SHARING_VIOLATION { return 0x20 }
 sub _NUMBER_OF_MCOOKIE_BYTES       { return 16 }
 sub _MAX_DISPLAY_LENGTH            { return 10 }
+sub _NUMBER_OF_TERM_ATTEMPTS       { return 4 }
 
 sub new {
     my ( $class, %parameters ) = @_;
@@ -383,48 +382,138 @@ sub _binary {
     if ( $self->{firefox_binary} ) {
         $binary = $self->{firefox_binary};
     }
-    if ( $OSNAME eq 'MSWin32' ) {
-        my $program_files_key;
-        foreach my $possible ( 'ProgramFiles(x86)', 'ProgramFiles' ) {
-            if ( $ENV{$possible} ) {
-                $program_files_key = $possible;
-                last;
+    else {
+        if ( $OSNAME eq 'MSWin32' ) {
+            my $program_files_key;
+            foreach my $possible ( 'ProgramFiles(x86)', 'ProgramFiles' ) {
+                if ( $ENV{$possible} ) {
+                    $program_files_key = $possible;
+                    last;
+                }
             }
+            $binary = File::Spec->catfile(
+                $ENV{$program_files_key},
+                'Mozilla Firefox',
+                'firefox.exe'
+            );
         }
-        $binary = File::Spec->catfile(
-            $ENV{$program_files_key},
-            'Mozilla Firefox',
-            'firefox.exe'
-        );
-    }
-    if ( $OSNAME eq 'darwin' ) {
-        $binary = '/Applications/Firefox.app/Contents/MacOS/firefox';
-    }
-    if ( $OSNAME eq 'cygwin' ) {
-        if ( -e "$ENV{PROGRAMFILES} (x86)" ) {
-            $binary = "$ENV{PROGRAMFILES} (x86)/Mozilla Firefox/firefox.exe";
+        elsif ( $OSNAME eq 'darwin' ) {
+            $binary = '/Applications/Firefox.app/Contents/MacOS/firefox';
         }
-        else {
-            $binary = "$ENV{PROGRAMFILES}/Mozilla Firefox/firefox.exe";
+        elsif ( $OSNAME eq 'cygwin' ) {
+            if ( -e "$ENV{PROGRAMFILES} (x86)" ) {
+                $binary =
+                  "$ENV{PROGRAMFILES} (x86)/Mozilla Firefox/firefox.exe";
+            }
+            else {
+                $binary = "$ENV{PROGRAMFILES}/Mozilla Firefox/firefox.exe";
+            }
         }
     }
     return $binary;
 }
 
-sub _firefox_alive {
+sub child_error {
+    my ($self) = @_;
+    return $self->{_child_error};
+}
+
+sub _signal_name {
+    my ( $self, $number ) = @_;
+    my @sig_names = split q[ ], $Config{sig_name};
+    return $sig_names[$number];
+}
+
+sub error_message {
+    my ($self) = @_;
+    my $message;
+    my $child_error = $self->child_error();
+    if ( !defined $self->child_error() ) {
+    }
+    elsif ( $OSNAME eq 'MSWin32' ) {
+    }
+    else {
+
+        if (   ( POSIX::WIFEXITED($child_error) )
+            || ( POSIX::WIFSIGNALED($child_error) ) )
+        {
+            if ( POSIX::WIFEXITED($child_error) ) {
+                if ( $child_error == 0 ) {
+                }
+                else {
+                    $message =
+                      'Exited with ' . POSIX::WEXITSTATUS($child_error);
+                }
+            }
+            elsif ( POSIX::WIFSIGNALED($child_error) ) {
+                my $name = $self->_signal_name( POSIX::WTERMSIG($child_error) );
+                if ( defined $name ) {
+                    $message = "Killed by $name signal ("
+                      . POSIX::WTERMSIG($child_error) . q[)];
+                }
+                else {
+                    $message = 'Killed by signal ('
+                      . POSIX::WTERMSIG($child_error) . q[)];
+                }
+            }
+        }
+        else {
+            if ( POSIX::WIFSTOPPED($child_error) ) {
+                my $name = $self->_signal_name( POSIX::WTERMSIG($child_error) );
+                if ( defined $name ) {
+                    $message = "Stopped by $name signal ("
+                      . POSIX::WSTOPSIG($child_error) . q[)];
+                }
+                else {
+                    $message = 'Stopped by signal ('
+                      . POSIX::WSTOPSIG($child_error) . q[)];
+                }
+            }
+            elsif ( POSIX::WIFCONTINUED($child_error) ) {
+                $message = 'Continuing';
+            }
+        }
+    }
+    return $message;
+}
+
+sub _reap {
     my ($self) = @_;
     if ( $OSNAME eq 'MSWin32' ) {
-        my $alive         = 0;
-        my $quoted_binary = quotemeta $self->_binary();
-        my $info          = Win32::Process::Info->new();
-        foreach my $process ( $info->GetProcInfo( $self->_pid() ) ) {
-            $alive = 1;
+        if ( $self->{_win32_process} ) {
+            $self->{_win32_process}->GetExitCode( my $exit_code );
+            if ( $exit_code != Win32::Process::STILL_ACTIVE() ) {
+                $self->{_child_error} = $exit_code;
+            }
         }
-        return $alive;
+    }
+    else {
+        while ( ( my $pid = waitpid _ANYPROCESS(), POSIX::WNOHANG() ) > 0 ) {
+            if ( $pid == $self->_pid() ) {
+                $self->{_child_error} = $CHILD_ERROR;
+            }
+            elsif ( ( $self->xvfb() ) && ( $pid == $self->xvfb() ) ) {
+                $self->{_xvfb_child_error} = $CHILD_ERROR;
+            }
+        }
+    }
+    return;
+}
+
+sub alive {
+    my ($self) = @_;
+    if ( $OSNAME eq 'MSWin32' ) {
+        if ( $self->{_win32_process} ) {
+            $self->{_win32_process}->GetExitCode( my $exit_code );
+            $self->_reap();
+            if ( $exit_code == Win32::Process::STILL_ACTIVE() ) {
+                return 1;
+            }
+        }
+        return 0;
     }
     elsif ( $self->_pid() ) {
-        waitpid _ANYPROCESS(), POSIX::WNOHANG();
-        waitpid $self->_pid(), POSIX::WNOHANG();
+        $self->_reap();
         return kill 0, $self->_pid();
     }
     else {
@@ -439,7 +528,7 @@ sub _setup_local_connection_to_firefox {
     my $binary = $self->_binary();
     my $socket;
     my $connected;
-    while ( ( !$connected ) && ( $self->_firefox_alive() ) ) {
+    while ( ( !$connected ) && ( $self->alive() ) ) {
         $socket = undef;
         socket $socket, Socket::PF_INET(), Socket::SOCK_STREAM(), 0
           or Carp::croak("Failed to create a socket: $EXTENDED_OS_ERROR");
@@ -462,13 +551,15 @@ sub _setup_local_connection_to_firefox {
                 "Failed to connect to $host on port $port:$EXTENDED_OS_ERROR");
         }
     }
-    waitpid _ANYPROCESS(), POSIX::WNOHANG();
-    waitpid $self->_pid(), POSIX::WNOHANG();
+    $self->_reap();
     if ( ( kill 0, $self->_pid() ) && ($socket) ) {
     }
     else {
-        Carp::croak( q[Firefox failed to start as '] . join q[ ],
-            $binary, @arguments );
+        my $error_message =
+          $self->error_message() ? q[:] . $self->error_message() : q[];
+        Carp::croak( q[Firefox failed to start as ']
+              . ( join q[ ], $binary, @arguments )
+              . $error_message );
     }
     return $socket;
 }
@@ -557,6 +648,10 @@ sub new_session {
 
 sub _create_capabilities {
     my ( $self, $parameters ) = @_;
+    my $pid = $parameters->{'moz:processID'};
+    if ( ($pid) && ( $OSNAME eq 'cygwin' ) ) {
+        $pid = $self->_pid();
+    }
     return Firefox::Marionette::Capabilities->new(
         accept_insecure_certs => $parameters->{acceptInsecureCerts} ? 1 : 0,
         page_load_strategy    => $parameters->{pageLoadStrategy},
@@ -571,7 +666,7 @@ sub _create_capabilities {
         platform_version         => $parameters->{platformVersion},
         moz_profile              => $parameters->{'moz:profile'},
         moz_webdriver_click      => $parameters->{'moz:webdriverClick'} ? 1 : 0,
-        moz_process_id           => $parameters->{'moz:processID'},
+        moz_process_id           => $pid,
         browser_name             => $parameters->{browserName},
         moz_headless             => $parameters->{'moz:headless'} ? 1 : 0,
         moz_accessibility_checks => $parameters->{'moz:accessibilityChecks'}
@@ -1252,7 +1347,15 @@ sub title {
 
 sub quit {
     my ( $self, $flags ) = @_;
-    if ( $self->_socket() ) {
+    if ( !$self->alive() ) {
+        my $socket = delete $self->{_socket};
+        if ($socket) {
+            close $socket
+              or Carp::croak(
+                "Failed to close socket to firefox:$EXTENDED_OS_ERROR");
+        }
+    }
+    elsif ( $self->_socket() ) {
         $flags ||=
           ['eAttemptQuit'];    # ["eConsiderQuit", "eAttemptQuit", "eForceQuit"]
         my $message_id = $self->_new_message_id();
@@ -1266,35 +1369,65 @@ sub quit {
           Carp::croak("Failed to close socket to firefox:$EXTENDED_OS_ERROR");
         if ( $OSNAME eq 'MSWin32' ) {
             $self->{_win32_process}->Wait( Win32::Process::INFINITE() );
+            $self->_reap();
         }
         else {
             while ( kill 0, $self->_pid() ) {
-                waitpid _ANYPROCESS(), POSIX::WNOHANG();
-            }
-        }
-    }
-    if ( ( $OSNAME eq 'MSWin32' ) or ( $OSNAME ne 'darwin' ) ) {
-    }
-    elsif ( kill 0, $self->_pid() ) {
-        my $kill_signal = $self->_signal_number('KILL');
-        if ( $kill_signal > 0 ) {
-            while ( kill $kill_signal, $self->_pid() ) {
-                waitpid _ANYPROCESS(), POSIX::WNOHANG();
+                $self->_reap();
             }
         }
     }
+    $self->_terminate_process();
     if ( my $pid = $self->xvfb() ) {
         my $int_signal = $self->_signal_number('INT');
         while ( kill 0, $pid ) {
             kill $int_signal, $pid;
             sleep 1;
-            waitpid _ANYPROCESS(), POSIX::WNOHANG();
+            $self->_reap();
         }
         delete $self->{_xvfb_display_number};
         delete $self->{_xvfb_authority_directory};
         delete $self->{_xvfb_pid};
     }
-    return $self;
+    return $self->child_error();
+}
+
+sub _terminate_process {
+    my ($self) = @_;
+    if ( $OSNAME eq 'MSWin32' ) {
+        if ( $self->{_win32_process} ) {
+            $self->{_win32_process}->Kill(1);
+            sleep 1;
+            $self->{_win32_process}->GetExitCode( my $exit_code );
+            while ( $exit_code == Win32::Process::STILL_ACTIVE() ) {
+                $self->{_win32_process}->Kill(1);
+                sleep 1;
+                $exit_code = $self->{_win32_process}->Kill(1);
+            }
+        }
+    }
+    elsif ( kill 0, $self->_pid() ) {
+        my $term_signal = $self->_signal_number('TERM')
+          ;    # https://support.mozilla.org/en-US/questions/752748
+        if ( $term_signal > 0 ) {
+            my $count = 0;
+            while (( $count < _NUMBER_OF_TERM_ATTEMPTS() )
+                && ( kill $term_signal, $self->_pid() ) )
+            {
+                $count += 1;
+                sleep 1;
+                $self->_reap();
+            }
+        }
+        my $kill_signal = $self->_signal_number('KILL');   # no more mr nice guy
+        if ( $kill_signal > 0 ) {
+            while ( kill $kill_signal, $self->_pid() ) {
+                sleep 1;
+                $self->_reap();
+            }
+        }
+    }
+    return;
 }
 
 sub context {
@@ -1594,7 +1727,7 @@ sub _read_from_socket {
     my ($self) = @_;
     my $number_of_bytes_in_response;
     my $initial_buffer;
-    while ( !defined $number_of_bytes_in_response ) {
+    while ( ( !defined $number_of_bytes_in_response ) && ( $self->alive() ) ) {
         my $number_of_bytes = sysread $self->_socket(), my $octet, 1;
         if ( defined $number_of_bytes ) {
             $initial_buffer .= $octet;
@@ -1605,7 +1738,10 @@ sub _read_from_socket {
     }
     my $number_of_bytes_already_read = 0;
     my $json                         = q[];
-    while ( $number_of_bytes_already_read < $number_of_bytes_in_response ) {
+    while (( defined $number_of_bytes_in_response )
+        && ( $number_of_bytes_already_read < $number_of_bytes_in_response )
+        && ( $self->alive() ) )
+    {
         my $number_of_bytes_read = sysread $self->_socket(), my $buffer,
           $number_of_bytes_in_response - $number_of_bytes_already_read;
         if ( defined $number_of_bytes_read ) {
@@ -1665,7 +1801,7 @@ Firefox::Marionette - Automate the Firefox browser with the Marionette protocol
 
 =head1 VERSION
 
-Version 0.23
+Version 0.24
 
 =head1 SYNOPSIS
 
@@ -2004,7 +2140,19 @@ returns the pid of the xvfb process if it exists.
 
 =head2 quit
 
-Marionette will stop accepting new connections before ending the current session, and finally attempting to quit the application
+Marionette will stop accepting new connections before ending the current session, and finally attempting to quit the application.  This method returns the $? (CHILD_ERROR) value for the Firefox process
+
+=head2 child_error
+
+This method returns the $? (CHILD_ERROR) for the Firefox process, or undefined if the process has not yet exited.
+
+=head2 error_message
+
+This method returns a human readable error message describing how the Firefox process exited (assuming it started okay).  On Win32 platforms this information is restricted to exit code.
+
+=head2 alive
+
+This method returns true or false depending on if the Firefox process is still running.
 
 =head1 DIAGNOSTICS
 
