@@ -14,7 +14,7 @@ BEGIN {
         require Win32;
     }
 }
-our $VERSION = '0.23';
+our $VERSION = '0.24';
 
 sub _ANY_PORT           { return 0 }
 sub _GETPWUID_DIR_INDEX { return 7 }
@@ -35,6 +35,10 @@ sub _profile_ini_directory {
           File::Spec->catdir( Win32::GetFolderPath( Win32::CSIDL_APPDATA() ),
             'Mozilla', 'Firefox' );
     }
+    elsif ( $OSNAME eq 'cygwin' ) {
+        $profile_ini_directory =
+          File::Spec->catdir( $ENV{APPDATA}, 'Mozilla', 'Firefox' );
+    }
     else {
         my $home_directory =
           ( getpwuid $EFFECTIVE_USER_ID )[ _GETPWUID_DIR_INDEX() ];
@@ -237,7 +241,7 @@ Firefox::Marionette::Profile - Represents a prefs.js Firefox Profile
 
 =head1 VERSION
 
-Version 0.23
+Version 0.24
 
 =head1 SYNOPSIS
 
